//File includes
#include <stdio.h>
#include "boards.h"
#include "app_util_platform.h"
#include "app_error.h"
#include "nrf_drv_twi.h"
#include "nrf_delay.h"
#include "nrf_log.h"
#include "nrf_log_ctrl.h"
#include "nrf_log_default_backends.h"
#include "nrf_gpio.h"

//Define the TWI instance
#define TWI_INSTANCE_ID     0

//Slave address and flow register address
#define SLAVE_ADDRESS       0x45U
#define FLOW_REG_ADDR       0x00U
#define TEMP_REG_ADDR       0x01U

//Flag to indicate if operation on TWI has ended
static volatile bool m_xfer_done = true;

//Define the TWI instance
static const nrf_drv_twi_t m_twi = NRF_DRV_TWI_INSTANCE(TWI_INSTANCE_ID);

//Buffer for samples
static uint8_t m_sample[2] = {0x0U, 0x0U};

//Function to handle the data that are produced at the TWI event
__STATIC_INLINE void data_handler(void)
{
  //Create a float variable to hold the compensated temperature value
  float temp_reading = 0.0f;

  //Create a variable to concatenate the two raw temperature bytes
  uint16_t raw_rdg = 0x0U;

  //Log the register values in the terminal
  raw_rdg = ((uint16_t)m_sample[0] << 8) | ((uint16_t)m_sample[1] & 0xFF);

  //Calculate the compensated temperature value by using the scaling function
  temp_reading = (float)raw_rdg * 50 / 65535;
    
  //Log the temperature measurement
  NRF_LOG_INFO("Temperature = " NRF_LOG_FLOAT_MARKER, NRF_LOG_FLOAT(temp_reading));

  //NRF_LOG_INFO("Register 0: %X ", m_sample[0]);
  //NRF_LOG_INFO("Register 1: %X ", m_sample[1]);
}

//Function to handle the TWI events
void twi_handler(nrf_drv_twi_evt_t const * p_event, void * p_context)
{
  switch (p_event->type)
  {
    case NRF_DRV_TWI_EVT_DONE:
      //In case the TWI event is the completion of the transfer
      if (p_event->xfer_desc.type == NRF_DRV_TWI_XFER_RX)
      {
        //Log the registers on the terminal
        data_handler();
      }
      //Clear the transfer flag
      m_xfer_done = true;
      break;

    default:
      break;
  }
}

//Initialize the TWI module
void twi_init (void)
{
  //Declare an error code
  ret_code_t err_code;

  //Initialize the configuration structure of the TWI
  const nrf_drv_twi_config_t twi_config = {
   .scl                = ARDUINO_SCL_PIN,
   .sda                = ARDUINO_SDA_PIN,
   .frequency          = NRF_DRV_TWI_FREQ_100K,
   .interrupt_priority = APP_IRQ_PRIORITY_HIGH,
   .clear_bus_init     = true
  };

  //Initialize the TWI module
  err_code = nrf_drv_twi_init(&m_twi, &twi_config, twi_handler, NULL);
  APP_ERROR_CHECK(err_code);

  //Enable the TWI module
  nrf_drv_twi_enable(&m_twi);
}

//Function to read data from the sensor
static void read_sensor_data()
{
  //Set flag to false, before the data transmission
  m_xfer_done = false;

  //Read two bytes from the sensor
  ret_code_t err_code = nrf_drv_twi_rx(&m_twi, SLAVE_ADDRESS, &m_sample, sizeof(m_sample));
  APP_ERROR_CHECK(err_code);
  while(m_xfer_done == false)
  {
    // Do nothing or sleep
  }
}

//Main function
int main(void)
{
  ret_code_t err_code=NRF_SUCCESS;

  //Initialize the nrf-log module
  APP_ERROR_CHECK(NRF_LOG_INIT(NULL));
  NRF_LOG_DEFAULT_BACKENDS_INIT();

  //Log a message to denote the beginning of the example
  NRF_LOG_INFO("\r\nTWI sensor example started.");
  NRF_LOG_FLUSH();

  nrf_delay_ms(100);
  //Initialize the TWI module
  twi_init();

  uint8_t temp_register[1] = {TEMP_REG_ADDR};
  uint8_t flow_register[1] = {FLOW_REG_ADDR};

  while (true)
  {
    //Repeat every 10ms
    nrf_delay_ms(10);
    do
    {
      __WFE();
    }while (m_xfer_done == false);

    //Set the twi transfer flag to false before a new transfer (write)
    m_xfer_done = false;

    //Change the register to read the temperature register
    err_code = nrf_drv_twi_tx(&m_twi, SLAVE_ADDRESS, temp_register, sizeof(temp_register), false);
    APP_ERROR_CHECK(err_code);
    while(m_xfer_done == false)
    {
      // Do nothing or sleep
    }

    //Read the temperature measurement
    read_sensor_data();
    NRF_LOG_FLUSH();

    //Set the twi transfer flag to false before a new transfer (write)
    m_xfer_done == false;

    //Change the register to read the flow register
    err_code = nrf_drv_twi_tx(&m_twi, SLAVE_ADDRESS, flow_register, sizeof(flow_register), false);
    APP_ERROR_CHECK(err_code);
    while(m_xfer_done == false)
    {
      // Do nothing or sleep
    }

    //Read the flow measurement
    read_sensor_data();
    NRF_LOG_FLUSH();

  }
}